'use strict';

var _classCallCheck = require('babel-runtime/helpers/class-call-check').default;

var _interopRequireDefault = require('babel-runtime/helpers/interop-require-default').default;

exports.__esModule = true;

var _arrayFind = require('array-find');

var _arrayFind2 = _interopRequireDefault(_arrayFind);

var _pluginHost = require('./plugin-host');

var _pluginHost2 = _interopRequireDefault(_pluginHost);

var Reporter = (function () {
    function Reporter(plugin, task, outStream) {
        _classCallCheck(this, Reporter);

        this.plugin = new _pluginHost2.default(plugin, outStream);

        this.passed = 0;
        this.total = task.tests.length;
        this.reportQueue = Reporter._createReportQueue(task);

        this._assignTaskEventHandlers(task);
    }

    // Static

    Reporter._createReportQueue = function _createReportQueue(task) {
        var runsPerTest = task.browserConnections.length;

        return task.tests.map(function (test) {
            return Reporter._createReportItem(test, runsPerTest);
        });
    };

    Reporter._createReportItem = function _createReportItem(test, runsPerTest) {
        return {
            fixtureName: test.fixture.name,
            fixturePath: test.fixture.path,
            testName: test.name,
            screenshotPath: null,
            pendingRuns: runsPerTest,
            errs: [],
            unstable: false,
            startTime: null
        };
    };

    Reporter._errorSorter = function _errorSorter(errs) {
        errs.sort(function (err1, err2) {
            err1 = err1.userAgent + err1.type;
            err2 = err2.userAgent + err2.type;

            if (err1 > err2) return 1;
            if (err1 < err2) return -1;
            return 0;
        });
    };

    Reporter.prototype._getReportItemForTestRun = function _getReportItemForTestRun(testRun) {
        var test = testRun.test;

        return _arrayFind2.default(this.reportQueue, function (i) {
            return i.fixturePath === test.fixture.path && i.testName === test.name;
        });
    };

    Reporter.prototype._shiftReportQueue = function _shiftReportQueue() {
        // NOTE: a completed report item is always at the front of the queue.
        // This happens because the next test can't be completed until the
        // previous one frees all browser connections.
        // Therefore, tests always get completed sequentially.
        var reportItem = this.reportQueue.shift();
        var durationMs = new Date() - reportItem.startTime;

        if (!reportItem.errs.length) this.passed++;

        Reporter._errorSorter(reportItem.errs);

        this.plugin.reportTestDone(reportItem.testName, reportItem.errs, durationMs, reportItem.unstable, reportItem.screenshotPath);

        // NOTE: here we assume that tests are sorted by fixture.
        // Therefore, if the next report item has a different
        // fixture, we can report this fixture start.
        var nextReportItem = this.reportQueue[0];

        if (nextReportItem && nextReportItem.fixturePath !== reportItem.fixturePath) this.plugin.reportFixtureStart(nextReportItem.fixtureName, nextReportItem.fixturePath);
    };

    Reporter.prototype._assignTaskEventHandlers = function _assignTaskEventHandlers(task) {
        var _this = this;

        task.once('start', function () {
            var startTime = new Date();
            var userAgents = task.browserConnections.map(function (bc) {
                return bc.userAgent;
            });
            var first = _this.reportQueue[0];

            _this.plugin.reportTaskStart(startTime, userAgents);
            _this.plugin.reportFixtureStart(first.fixtureName, first.fixturePath);
        });

        task.on('test-run-start', function (testRun) {
            var reportItem = _this._getReportItemForTestRun(testRun);

            if (!reportItem.startTime) reportItem.startTime = new Date();
        });

        task.on('test-run-done', function (testRun) {
            var reportItem = _this._getReportItemForTestRun(testRun);
            var userAgent = testRun.browserConnection.userAgent;

            testRun.errs.forEach(function (err) {
                return err.userAgent = userAgent;
            });

            reportItem.pendingRuns--;
            reportItem.errs = reportItem.errs.concat(testRun.errs);
            reportItem.unstable = reportItem.unstable || testRun.unstable;

            if (!reportItem.pendingRuns) {
                if (task.screenshots.hasCapturedFor(testRun.test)) reportItem.screenshotPath = task.screenshots.getPathFor(testRun.test);

                _this._shiftReportQueue();
            }
        });

        task.once('done', function () {
            var endTime = new Date();

            _this.plugin.reportTaskDone(endTime, _this.passed, _this.total);
        });
    };

    return Reporter;
})();

exports.default = Reporter;
module.exports = exports.default;